# include "minimize.h"

/* max permitted O-O distance */
#define OO 5.5

/* H-bond limit */
#define HO 2.5

static void nearestdr(vector dr)
{
  int k;
  loop (k,0,3) {
    if (dr[k]>box.Lh[k]) dr[k]-=box.L[k];
    else if (dr[k]<-box.Lh[k]) dr[k]+=box.L[k]; }
}

static struct bjerrum_t {
  int D1[2],D2[2]; /* [0]=D-defect, [1]=L-defect; D1,D2=both molecules */
} bjerrum;

static double maxm=0,minm=9e9;
static int iO,iH1,iH2;
static int (*hb)[2];

void bjerrum_read_oo(void) /******************************** bjerrum_read_oo */
{
  int ns,i;
  molecule_t *m1;
  siteinfo_t *si1;

  loop (i,0,2) bjerrum.D1[i]=-1,bjerrum.D2[i]=-1;

  if (maxm==0) {
    FILE *oo;
    char line[64];
    int n=0,a,b;

    allocarrayzero(hb,No.N*2);

    /* O,H sites */
    m1=molec;
    ns=m1->ns;
    si1=spec[m1->sp]->si;
    loop (i,0,ns) {
      if (si1[i].mass>maxm) {
        maxm=si1[i].mass; iO=i; }
      if (si1[i].mass>0 && si1[i].mass<minm) {
        minm=si1[i].mass*1.0001;
        iH1=iH2,iH2=i; } }
    prt("molecule 0 sites: O=%d H=%d,%d",iO,iH1,iH2);

    oo=fopen(Fn("oo"),"rt");
    if (!oo) ERROR(("bjerrum: neighbor list (H-bonded water-water pairs) %s missing",lastFn))

    while (fgets(line,64,oo)) {
      if (n>=2*No.N) ERROR(("%s: too many lines",lastFn))
      sscanf(line,"%d%d",&a,&b);
      if (a<0 || a>=No.N || b<0 || b>=No.N) ERROR(("%s: water molecule numbers out of range"))

      hb[n][0]=a;
      hb[n][1]=b;
      if (strchr(line,'D')) bjerrum.D1[0]=a,bjerrum.D2[0]=b;
      if (strchr(line,'L')) bjerrum.D1[1]=a,bjerrum.D2[1]=b;
      n++; }

    if (n<2*No.N) ERROR(("%s: too few lines (called because bj.mode&1)",lastFn))
    prt("%s read: D-defect: D1=%d D2=%d, L-defect: L1=%d L2=%d",
        lastFn,bjerrum.D1[0],bjerrum.D2[0],bjerrum.D1[1],bjerrum.D2[1]); }
}

void bjerrum_topology(ToIntPtr A) /************************ bjerrum_topology */
{
  int n,n1,n2,ns;
  molecule_t *m1;
  siteinfo_t *si1;
  vector *r1,*r2,dr;
  int nHB1,nHB2,nL,nL1,nL2,nD,nD1,nD2;

  /* finding a defect - topology way */
  nL=nD=0;
  prt("bjerrum: t=%f",t);

  if (!hb) ERROR(("no hb: bjerrum_read_oo should be called first"))

  loop (n,0,2*No.N) {
    n1=hb[n][0];
    n2=hb[n][1];
    m1=molec+n1;
    si1=spec[m1->sp]->si;
    r1=rof(m1,A->rp);
    r2=rof(molec+n2,A->rp);

    VVV(dr,=r1[iO],-r2[iO]) nearestdr(dr);

    if (SQR(dr)>Sqr(OO)) prt("O%d-O%d=%g exceeded the limit (%g)",n1,n2,sqrt(SQR(dr)),(double)OO);

    nHB1=0;
    VVV(dr,=r1[iO],-r2[iH1]) nearestdr(dr); if (SQR(dr)<Sqr(HO)) nHB1++;
    VVV(dr,=r1[iO],-r2[iH2]) nearestdr(dr); if (SQR(dr)<Sqr(HO)) nHB1++;

    nHB2=0;
    VVV(dr,=r2[iO],-r1[iH1]) nearestdr(dr); if (SQR(dr)<Sqr(HO)) nHB2++;
    VVV(dr,=r2[iO],-r1[iH2]) nearestdr(dr); if (SQR(dr)<Sqr(HO)) nHB2++;

    if (nHB1==0 && nHB2==0) {
      /* L-defect */
      nL++; nL1=n1; nL2=n2; }

    if (nHB1==1 && nHB2==1) {
      /* D-defect */
      nD++; nD1=n1; nD2=n2; } }

  prt_("!nD=%d nD1=%d nD2=%d   ",nD,nD1,nD2);
  prt("nL=%d nL1=%d nL2=%d BJERRUM1",nL,nL1,nL2);
}

static struct bjerrumminimize_s {
  ToIntPtr A;
  double sig;
  int it;
} bjmin;

double SumGauss(vector R)
{
  double sumq=0,sigq=Sqr(bjmin.sig)*2,rr;
  int n,i;
  molecule_t *m;
  siteinfo_t *si;
  vector *r,dr;
  
  bjmin.it++;
  
  loop (n,0,No.N) {
    m=molec+n;
    si=spec[m->sp]->si;
    r=rof(m,bjmin.A->rp);
    loop (i,0,m->ns) if (si[i].charge) {
#  if 0
      /* consider 27 images: -1,0,+1 SLOW! */
      int j[3];

      loopto (j[0],-1,1)
        loopto (j[1],-1,1)
          loopto (j[2],-1,1) {
            VVVVV(dr,=R,-r[i],+j,*box.L)
            rr=SQR(dr);
            sumq+=exp(-rr/sigq)*si[i].charge; }
#  else
            /* nearest-image */
            VVV(dr,=R,-r[i])
            nearestdr(dr);
            rr=SQR(dr);
            sumq+=exp(-rr/sigq)*si[i].charge;
#  endif
    } }

  if (bjmin.sig>0) return sumq;
  else return -sumq;
}

void bjerrum_gauss(ToIntPtr A,double sig0,double eps) /******* bjerrum_gauss */
{
  int n,i,k,idefect,nit;
  molecule_t *m;
  siteinfo_t *si;
  vector *r,dr,R0,R;
  double sig,d,sigq,sumq,rr;
  double d0=cbrt(PROD(box.L))*0.25;

  bjmin.A=A;

  loop (idefect,0,2) {
    /* idefect=0: D-defect, idefect=1: L-defect */

    m=molec+bjerrum.D1[idefect];
    r=rof(m,A->rp);
    VO(R0,=0)
    loop (i,0,m->ns) VV(R0,+=r[i])
    VO(R0,/=m->ns)

    m=molec+bjerrum.D2[idefect];
    r=rof(m,A->rp);
    VO(R,=0)
    loop (i,0,m->ns) VV(R,+=r[i])
    VO(R,/=m->ns)

    /* CoM of two molecules comprising a defect */
    loop (k,0,DIM) {
      if (R0[k]>R[k]+box.Lh[k]) R0[k]-=box.L[k];
      if (R0[k]<R[k]-box.Lh[k]) R0[k]+=box.L[k];
      R0[k]=(R0[k]+R[k])/2;
      if (R0[k]<0) R0[k]+=box.L[k];
      if (R0[k]>box.L[k]) R0[k]-=box.L[k]; }

    for (sig=sig0; sig<d0; sig*=sqrt(sqrt(2))) {

      VV(R,=R0)

      bjmin.sig=idefect?-sig:sig;
      bjmin.sig=-bjmin.sig;
      bjmin.it=0;

      Minimize(MINIMIZE_AMOEBA,3,R,SumGauss,1000,eps,bjmin.sig*0.01,2);
      //      Minimize(MINIMIZE_NR,3,R,SumGauss,100,eps,bjmin.sig*0.01,2);

      sumq=SumGauss(R)/electron;

      prt("%9.6f %9.6f [%10.6f %10.6f %10.6f] %d BJERRUM2",
          sig,idefect?-sumq:sumq,VARG(R0),bjmin.it); }
  }
}
